src/java.base/share/classes/java/lang/Shutdown.java
changeset 48942 a6c4b85163c1
parent 47216 71c04702a3d5
child 49186 b8bfadfacded
--- a/src/java.base/share/classes/java/lang/Shutdown.java	Fri Feb 23 14:26:29 2018 -0500
+++ b/src/java.base/share/classes/java/lang/Shutdown.java	Fri Feb 23 12:10:56 2018 -0800
@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1999, 2005, Oracle and/or its affiliates. All rights reserved.
+ * Copyright (c) 1999, 2018, Oracle and/or its affiliates. All rights reserved.
  * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
  *
  * This code is free software; you can redistribute it and/or modify it
@@ -26,35 +26,33 @@
 package java.lang;
 
 
+import jdk.internal.misc.VM;
+
 /**
  * Package-private utility class containing data structures and logic
  * governing the virtual-machine shutdown sequence.
  *
  * @author   Mark Reinhold
  * @since    1.3
+ *
+ * @see java.io.Console
+ * @see ApplicationShutdownHooks
+ * @see java.io.DeleteOnExitHook
  */
 
 class Shutdown {
 
-    /* Shutdown state */
-    private static final int RUNNING = 0;
-    private static final int HOOKS = 1;
-    private static final int FINALIZERS = 2;
-    private static int state = RUNNING;
-
-    /* Should we run all finalizers upon exit? */
-    private static boolean runFinalizersOnExit = false;
-
     // The system shutdown hooks are registered with a predefined slot.
     // The list of shutdown hooks is as follows:
     // (0) Console restore hook
-    // (1) Application hooks
+    // (1) ApplicationShutdownHooks that invokes all registered application
+    //     shutdown hooks and waits until they finish
     // (2) DeleteOnExit hook
     private static final int MAX_SYSTEM_HOOKS = 10;
     private static final Runnable[] hooks = new Runnable[MAX_SYSTEM_HOOKS];
 
     // the index of the currently running shutdown hook to the hooks array
-    private static int currentRunningHook = 0;
+    private static int currentRunningHook = -1;
 
     /* The preceding static fields are protected by this lock */
     private static class Lock { };
@@ -63,17 +61,9 @@
     /* Lock object for the native halt method */
     private static Object haltLock = new Lock();
 
-    /* Invoked by Runtime.runFinalizersOnExit */
-    static void setRunFinalizersOnExit(boolean run) {
-        synchronized (lock) {
-            runFinalizersOnExit = run;
-        }
-    }
-
-
     /**
      * Add a new shutdown hook.  Checks the shutdown state and the hook itself,
      * but does not do any security checks.
      *
      * The registerShutdownInProgress parameter should be false except
      * registering the DeleteOnExitHook since the first file may
@@ -92,15 +82,18 @@
      *         already passes the given slot
      */
     static void add(int slot, boolean registerShutdownInProgress, Runnable hook) {
+        if (slot < 0 || slot >= MAX_SYSTEM_HOOKS) {
+            throw new IllegalArgumentException("Invalid slot: " + slot);
+        }
         synchronized (lock) {
             if (hooks[slot] != null)
                 throw new InternalError("Shutdown hook at slot " + slot + " already registered");
 
             if (!registerShutdownInProgress) {
-                if (state > RUNNING)
+                if (currentRunningHook >= 0)
                     throw new IllegalStateException("Shutdown in progress");
             } else {
-                if (state > HOOKS || (state == HOOKS && slot <= currentRunningHook))
+                if (VM.isShutdown() || slot <= currentRunningHook)
                     throw new IllegalStateException("Shutdown in progress");
             }
 
@@ -108,9 +101,23 @@
         }
     }
 
-    /* Run all registered shutdown hooks
+    /* Run all system shutdown hooks.
+     *
+     * The system shutdown hooks are run in the thread synchronized on
+     * Shutdown.class.  Other threads calling Runtime::exit, Runtime::halt
+     * or JNI DestroyJavaVM will block indefinitely.
+     *
+     * ApplicationShutdownHooks is registered as one single hook that starts
+     * all application shutdown hooks and waits until they finish.
      */
     private static void runHooks() {
+        synchronized (lock) {
+            /* Guard against the possibility of a daemon thread invoking exit
+             * after DestroyJavaVM initiates the shutdown sequence
+             */
+            if (VM.isShutdown()) return;
+        }
+
         for (int i=0; i < MAX_SYSTEM_HOOKS; i++) {
             try {
                 Runnable hook;
@@ -121,13 +128,16 @@
                     hook = hooks[i];
                 }
                 if (hook != null) hook.run();
-            } catch(Throwable t) {
+            } catch (Throwable t) {
                 if (t instanceof ThreadDeath) {
                     ThreadDeath td = (ThreadDeath)t;
                     throw td;
                 }
             }
         }
+
+        // set shutdown state
+        VM.shutdown();
     }
 
     /* The halt method is synchronized on the halt lock
@@ -142,74 +152,22 @@
 
     static native void halt0(int status);
 
-    /* Wormhole for invoking java.lang.ref.Finalizer.runAllFinalizers */
-    private static native void runAllFinalizers();
-
-
-    /* The actual shutdown sequence is defined here.
-     *
-     * If it weren't for runFinalizersOnExit, this would be simple -- we'd just
-     * run the hooks and then halt.  Instead we need to keep track of whether
-     * we're running hooks or finalizers.  In the latter case a finalizer could
-     * invoke exit(1) to cause immediate termination, while in the former case
-     * any further invocations of exit(n), for any n, simply stall.  Note that
-     * if on-exit finalizers are enabled they're run iff the shutdown is
-     * initiated by an exit(0); they're never run on exit(n) for n != 0 or in
-     * response to SIGINT, SIGTERM, etc.
-     */
-    private static void sequence() {
-        synchronized (lock) {
-            /* Guard against the possibility of a daemon thread invoking exit
-             * after DestroyJavaVM initiates the shutdown sequence
-             */
-            if (state != HOOKS) return;
-        }
-        runHooks();
-        boolean rfoe;
-        synchronized (lock) {
-            state = FINALIZERS;
-            rfoe = runFinalizersOnExit;
-        }
-        if (rfoe) runAllFinalizers();
-    }
-
-
     /* Invoked by Runtime.exit, which does all the security checks.
      * Also invoked by handlers for system-provided termination events,
      * which should pass a nonzero status code.
      */
     static void exit(int status) {
-        boolean runMoreFinalizers = false;
         synchronized (lock) {
-            if (status != 0) runFinalizersOnExit = false;
-            switch (state) {
-            case RUNNING:       /* Initiate shutdown */
-                state = HOOKS;
-                break;
-            case HOOKS:         /* Stall and halt */
-                break;
-            case FINALIZERS:
-                if (status != 0) {
-                    /* Halt immediately on nonzero status */
-                    halt(status);
-                } else {
-                    /* Compatibility with old behavior:
-                     * Run more finalizers and then halt
-                     */
-                    runMoreFinalizers = runFinalizersOnExit;
-                }
-                break;
+            if (status != 0 && VM.isShutdown()) {
+                /* Halt immediately on nonzero status */
+                halt(status);
             }
         }
-        if (runMoreFinalizers) {
-            runAllFinalizers();
-            halt(status);
-        }
         synchronized (Shutdown.class) {
             /* Synchronize on the class object, causing any other thread
              * that attempts to initiate shutdown to stall indefinitely
              */
-            sequence();
+            runHooks();
             halt(status);
         }
     }
@@ -220,18 +178,8 @@
      * actually halt the VM.
      */
     static void shutdown() {
-        synchronized (lock) {
-            switch (state) {
-            case RUNNING:       /* Initiate shutdown */
-                state = HOOKS;
-                break;
-            case HOOKS:         /* Stall and then return */
-            case FINALIZERS:
-                break;
-            }
-        }
         synchronized (Shutdown.class) {
-            sequence();
+            runHooks();
         }
     }
@@ -63,17 +61,9 @@
     /**
-     * Add a new shutdown hook.  Checks the shutdown state and the hook itself,
-     * but does not do any security checks.
+     * Add a new system shutdown hook.  Checks the shutdown state and
+     * the hook itself, but does not do any security checks.
      *
      * The registerShutdownInProgress parameter should be false except